-- Asset Manager
-- Handles loading in assets in a specific format
-- as well as managing access to them

local AssetManager = {}
local AssetManagerConstructor = {}
local mt = {__index = AssetManager}

function AssetManagerConstructor.new(o)
    o = o or {}
    setmetatable(o, mt)
    o.images = {}
    o.sounds = {}
    o.music = {}
    o.fonts = {}
    o.tilesets = {}
    return o
end



-- Data that probably shouldn't be here
-- but hell if I can think where it should
-- otherwise go
local fontSizes = {}
fontSizes["aesymatt"] = {11, 14, 16, 28}


function AssetManagerConstructor.loadFromFolder(folderPath, filterType, loadedItemTable, loadedCallback, subDirectory, dontAddToTable)
    subDirectory = subDirectory or ""
    local filenames = love.filesystem.getDirectoryItems(folderPath)
    for _, foundFilename in pairs(filenames) do
        if love.filesystem.isDirectory(folderPath .. "/" .. foundFilename) then
            AssetManagerConstructor.loadFromFolder(folderPath .. "/" .. foundFilename .. "/", filterType, loadedItemTable, loadedCallback, foundFilename .. "/", dontAddToTable)
        end        
        local fileName, fileType = string.match(foundFilename, "(.+)%.(.+)$")
        if fileType == filterType or (fileType and filterType == "*") then -- don't load directories
            local loadedFile = { name = subDirectory .. fileName }
            if loadedCallback then                
                loadedCallback(loadedFile, folderPath .. fileName .. "." .. fileType, subDirectory)
            end
            loadedItemTable[#loadedItemTable + 1] = loadedFile
            loadedItemTable[loadedFile.name] = loadedFile
        end
    end
end

function AssetManager:loadAllAssets()    
    self:loadSounds()
    --self:loadShaders()
    self:loadImages()
    self:loadTilesets(TILE_SIZE)
    self:loadMusic()
    self:loadFonts()
end

--function AssetManager:loadShaders()
--    self.shaders = {}
--    AssetManagerConstructor.loadFromFolder("assets/shaders/", "glsl", self.shaders, function(loaded, filePath, subDirectory)
--            loaded.shader = love.graphics.newShader(filePath)
--            loaded.shader:send("resolution", { love.graphics.getWidth(), love.graphics.getHeight() })
--            print("loaded shader " .. loaded.name)
--        end)
--end

--function AssetManager:getShader(name)
--    if self.shaders[name] then
--        return self.shaders[name].shader
--    end
--    return nil
--end

function AssetManager:loadSounds()
    self.sounds = {}
    AssetManagerConstructor.loadFromFolder("assets/sounds/", "*", self.sounds, function(loaded, filePath, subDirectory)
            loaded.sound = love.audio.newSource(filePath, "static")
            print("loaded sound " .. loaded.name)
        end)
end

function AssetManager:getSound(name)
    if self.sounds[name] then
        return self.sounds[name].sound
    end
    return nil
end

function AssetManager:loadImages()
    self.images = {}
    AssetManagerConstructor.loadFromFolder("assets/images/", "png", self.images, function (loaded, filePath, subDirectory)
            loaded.texture = love.graphics.newImage(filePath)       
            print("loaded image " .. loaded.name)
        end) 
end

function AssetManager:getImage(name)
    if self.images[name] then
        return self.images[name].texture
    end
    return nil
end

function AssetManager:loadTilesets(tileSize)
    self.tilesets = {}
    AssetManagerConstructor.loadFromFolder("assets/tilesets/", "png", self.tilesets, function (loaded, filePath, subDirectory)
            local texture  = love.graphics.newImage(filePath)
            loaded.texture = texture
            
            local tilesetWidth = texture:getWidth()
            local tilesetHeight = texture:getHeight()
            local tileRows = math.floor(tilesetHeight / tileSize)
            local tileColumns = math.floor(tilesetWidth / tileSize)
            
            local tileQuads = {}
            for tileY = 0, tileRows do
                for tileX = 0, tileColumns do
                    local tileQuad = love.graphics.newQuad(tileX * tileSize, tileY * tileSize, tileSize, tileSize, tilesetWidth, tilesetHeight)
                    tileQuads[tileY * tileColumns + tileX + 1] = tileQuad -- tiles begin indexed at 1, not 0
                end
            end       
            loaded.quads = tileQuads
            
            if USE_NEAREST_NEIGHBOR_SCALING then
                loaded.texture:setFilter("nearest")
            end         
            print("loaded tileset " .. loaded.name)
        end) 
end

function AssetManager:getTileset(name)
    if self.tilesets[name] then
        return self.tilesets[name]
    end
    return nil
end

function AssetManager:loadMusic()
    self.music = {}
    AssetManagerConstructor.loadFromFolder("assets/music/", "ogg", self.music, function (loaded, filePath, subDirectory)
            loaded.musicData = love.audio.newSource(filePath)
            print("loaded music " .. loaded.name)
        end)
end

function AssetManager:getMusic(name)
    if self.music[name] then
        return self.music[name].musicData
    end
    return nil
end

function AssetManager:loadFonts()
    self.fonts = {}
    AssetManagerConstructor.loadFromFolder("assets/fonts/", "*", self.fonts, function (loaded, filePath, subDirectory)            
            local sizes = fontSizes[loaded.name]
            if sizes then
                for _, size in pairs(sizes) do
                    local fontName = loaded.name .. size
                    local fullLoaded = {name=fontName, size=size}
                    fullLoaded.font = love.graphics.newFont(filePath, size)                    
                    self.fonts[#self.fonts + 1] = fullLoaded
                    self.fonts[fontName] = fullLoaded
                    print("loaded font " .. fontName)
                end
            end
        end, "", true)
end


function AssetManager:getFont(name)
    if self.fonts[name] then
        return self.fonts[name].font
    end
    return nil
end


return AssetManagerConstructor