require 'globals'
local anim8 = require 'lib.anim8'

local Player = {}
local PlayerConstructor = {}
PlayerConstructor.mt = {__index = Player}

function PlayerConstructor.new(o, x, floorY, assets)
    o = o or {}
    setmetatable(o, PlayerConstructor.mt)

    o.assets = assets
    o.alive = true
    o.width = 20
    o.height = 40        

    o.x = x
    o.y = floorY - o.height/2

    o.vel = {x=0, y=0}
    o.groundAccel = 0
    o.totalGroundSpeed = 100
    o.groundSpeed = 100

    o.offset = {x=-15, y=-10}
    o.image = assets:getImage("player")
    local g = anim8.newGrid(50, 50, o.image:getWidth(), o.image:getHeight())    

    o.runAnimation = anim8.newAnimation(g('1-4', 1), 0.1)
    o.jumpAnimation = anim8.newAnimation(g('5-6', 1), 0.2)
    o.fallAnimation = anim8.newAnimation(g('7-8', 1), 0.2)
    o.diveAnimation = anim8.newAnimation(g('9-10', 1), 0.2)
    o.dieAnimation = anim8.newAnimation(g('11-14', 1), 0.2, "pauseAtEnd")
    o.currentAnimation = o.runAnimation

    o.jumpSound = assets:getSound("jumpPlayer")
    o.landSound = assets:getSound("land")
    o.diveSound = assets:getSound("dive")

    return o
end

function Player:update(dt, floorY, extraSpeed)
    self.currentAnimation:update(dt)

    self.x = self.x + self.vel.x * dt
    self.y = self.y + self.vel.y * dt
    self.vel.x = self.vel.x * 0.9
    self.vel.y = self.vel.y + GRAVITY



    -- put player back on floor if they've gone under it
    if(self.y + self.height/2 > floorY) then
        self.vel.y = 0
        self.y = floorY - self.height/2        
        if self.inAir then
            playSound(self.landSound, 0.3, {x = self.x, y = self.y})
            self.currentAnimation = self.runAnimation
        end
        self.inAir = false
    end

    if self.inAir then
        if self.vel.y < 0 then
            self.currentAnimation = self.jumpAnimation
        else
            if self.currentAnimation ~= self.diveAnimation then
                self.currentAnimation = self.fallAnimation
            end
        end
    else 
        self.currentAnimation = self.runAnimation
    end


    self.groundSpeed = self.groundSpeed + self.groundAccel * dt
    if self.groundSpeed < MIN_SPEED then
        self.groundSpeed = MIN_SPEED
    elseif self.groundSpeed > MAX_SPEED then
        self.groundSpeed = MAX_SPEED
    end
    self.totalGroundSpeed = self.groundSpeed + extraSpeed

    if not self.alive and not self.inAir then
        self.currentAnimation = self.dieAnimation
    end

end

function Player:draw()
    love.graphics.setColor(255, 255, 255)
    self.currentAnimation:draw(self.image, self.x - self.width/2 + self.offset.x, self.y - self.height/2 + self.offset.y)
end

function Player:jump()
    if not self.inAir then
        self.vel.y = -JUMP_STRENGTH
        self.inAir = true
        playSound(self.jumpSound, 0.3, {x = self.x, y = self.y})
    end
end

function Player:endJump()
    if self.inAir and self.vel.y < 0 then
        self.vel.y = self.vel.y / 4
    end
end

function Player:dive()
    if self.inAir then
        self.vel.y = self.vel.y + DIVE_STRENGTH
        playSound(self.diveSound, 0.3, {x = self.x, y = self.y})
        self.currentAnimation = self.diveAnimation
    end    
end

function Player:beginAccel(direction)
    if direction > 0 then
        self.groundAccel = RUN_ACCEL
    else
        self.groundAccel = -RUN_DECEL
    end
end

function Player:endAccel()
    self.groundAccel = 0
end



return PlayerConstructor