require 'globals'
local anim8 = require 'lib.anim8'
local FireParticles = require 'particle.fire'

local Settlement = {}
local SettlementConstructor = {}
SettlementConstructor.mt = {__index = Settlement}

function SettlementConstructor.new(o, assets)
    o = o or {}
    setmetatable(o, SettlementConstructor.mt)

    o.x = o.x or 0
    o.y = o.y or 0 -- y is the bottom part of the settlement, not middle
    o.alive = true
    o.assets = assets

    local scalingFactor = math.random(0.85, 0.95)
    o.width = SETTLEMENT_WIDTH * scalingFactor

    -- villager and torch animations

    o.villagerImage = assets:getImage("villager")
    o.torchImage = assets:getImage("torch")

    local g = anim8.newGrid(50, 50, o.villagerImage:getWidth(), o.villagerImage:getHeight())    
    o.walkAnimationL = anim8.newAnimation(g('1-4', 1), 0.1)
    o.jumpAnimationL = anim8.newAnimation(g('5-6', 1), 0.2)
    o.fallAnimationL = anim8.newAnimation(g('7-8', 1), 0.2)
    o.walkAnimationR = o.walkAnimationL:clone():flipH()
    o.jumpAnimationR = o.jumpAnimationL:clone():flipH()
    o.fallAnimationR = o.fallAnimationL:clone():flipH()
 
    g = anim8.newGrid(20, 70, o.torchImage:getWidth(), o.torchImage:getHeight())
    o.unlitAnimation = anim8.newAnimation(g(1, 1), 1)
    o.litAnimation = anim8.newAnimation(g('2-4', 1), 0.1)

    o.torches = {}
    local torchWidth = 20
    local torchHeight = 70
    local torchSpacing = (o.width - (torchWidth*SETTLEMENT_TORCH_COUNT)) / SETTLEMENT_TORCH_COUNT
    for i=0,SETTLEMENT_TORCH_COUNT-1 do
        o.torches[#o.torches+1] = {
            x = (i - SETTLEMENT_TORCH_COUNT/2) * (torchWidth + torchSpacing*0.5) + torchSpacing*0.5,
            width = torchWidth,
            height = torchHeight,
            lit = false,
            animation = o.unlitAnimation
        }
    end

    o.torchFlames = {}

    o.villagers = {}
    local villagerWidth = 50
    local villagerHeight = 50
    local numVillagers = love.math.random(SETTLEMENT_TORCH_COUNT/2, SETTLEMENT_TORCH_COUNT)
    for i=1,numVillagers do
        local wander = math.random(-50, 50)
        local animDirection = wander < 0 and "R" or "L"
        local walkAnimation = o["walkAnimation" .. animDirection]:clone()
        local jumpAnimation = o["jumpAnimation" .. animDirection]:clone()
        local fallAnimation = o["fallAnimation" .. animDirection]:clone()
        o.villagers[#o.villagers+1] = {
            x = math.random(-o.width/4, o.width/4),
            y = -villagerHeight/2,
            vy = 0,
            wanderX = wander,
            hop = math.random(100, 200),
            width = villagerWidth,
            height = villagerHeight,
            color = {r=128 + math.random(128), g=128 + math.random(128), b=128 + math.random(128)},            
            walkAnimation = walkAnimation,
            jumpAnimation = jumpAnimation,
            fallAnimation = fallAnimation,
            animation = walkAnimation
        }      
    end



    o.torchHitSound = assets:getSound("lightTorch")
    o.villagerJumpSound = assets:getSound("jumpVillager")
    o.villagerJumpSound:setVolume(0.5)
    o.villagerCheerSound = assets:getSound("cheer")
    o.villagerCheerSound:setVolume(0.8)

    return o
end

function Settlement:update(dt, movement, playerX)
    if self.alive then
        self.x = self.x - movement
        if self.x + self.width/2 < 0 then
            self.alive = false
        end

        -- update villagers
        for _,v in ipairs(self.villagers) do
            v.animation:update(dt)
            
            if v.inAir then
                if v.vy < 0 then
                    v.animation = v.jumpAnimation
                else
                    v.animation = v.fallAnimation
                end
            else
                v.animation = v.walkAnimation               
            end
            
            
            v.x = v.x + v.wanderX * dt
            if v.x < -SETTLEMENT_WIDTH/2 then
                v.x = -SETTLEMENT_WIDTH/2
            elseif v.x > SETTLEMENT_WIDTH/2 then
                v.x = SETTLEMENT_WIDTH/2
            end            

            if math.random() > 0.999 then
                playSound(self.villagerCheerSound, 1.2, {x = self.x + v.x, y = self.y + v.y})
            end

            if not v.inAir and math.random() > 0.99 then
                v.vy = -v.hop
                v.inAir = true
                playSound(self.villagerJumpSound, 0.5, {x = self.x + v.x, y = self.y + v.y})
            end
            v.vy = v.vy + GRAVITY
            v.y = v.y + v.vy * dt

            if v.y > -v.height/2 then
                v.y = -v.height/2
                v.vy = 0
                v.inAir = false
            end            
        end
    end

    for _,v in ipairs(self.torchFlames) do
        v:update(dt, movement)
        v.x = self.x + v.torchX
    end
    
    for _,v in ipairs(self.torches) do
        v.animation:update(dt)
    end    

end

function Settlement:checkTorches(playerX)
    local scoreBoost = 0
    for _,v in ipairs(self.torches) do
        if playerX > self.x + v.x and not v.lit then
            v.lit = true
            scoreBoost = scoreBoost + SETTLEMENT_TORCH_SCORE
            playSound(self.torchHitSound, 0.5, {x = self.x + v.x, y = self.y})
            local torchFlame = FireParticles.new({x = self.x + v.x, y = self.y - v.height + 10, r = 15, num = 20})
            self.torchFlames[#self.torchFlames+1] = torchFlame
            torchFlame.torchX = v.x
            v.animation = self.litAnimation
        end
    end   
    return scoreBoost
end



function Settlement:draw()
    if self.alive then        
        for _,v in ipairs(self.torches) do
            love.graphics.setColor(255, 255, 255)
            v.animation:draw(self.torchImage, self.x + v.x - v.width/2, self.y - v.height)
        end

        for _,v in ipairs(self.torchFlames) do
            v:draw()
        end

        for _,v in ipairs(self.villagers) do
            love.graphics.setColor(v.color.r, v.color.g, v.color.b)
            --love.graphics.rectangle("fill", self.x + v.x - v.width/2 + v.offset.x, self.y + v.y - v.height/2 + v.offset.y, v.width, v.height)
            v.animation:draw(self.villagerImage, self.x + v.x - v.width/2, self.y + v.y - v.height/2)
        end        


    end
end


return SettlementConstructor