require 'globals'
local StateConstructor = require 'state.state'
local Player = require 'player'
local Settlement = require 'settlement'
local SnowParticles = require 'particle.snow'
local FireParticles = require 'particle.fire'
local StatusBar = require 'gui.status-bar'

local PlayState = setmetatable({}, StateConstructor.mt)
local PlayStateConstructor = {}
local mt = {__index = PlayState}

local hazardTypes = { {
        name = "cold lump",
        color = {r=195, g=10, b=30},
        imageName = "rock",
        width = 40, 
        height = 30,
        place = "floor"
        }, {
        name = "glacial boulder",
        color = {r=195, g=10, b=30},
        imageName = "boulder",
        width = 80, 
        height = 50,
        place = "floor"
        }, {
        name = "hanging ice crystal",
        color = {r=195, g=10, b=60},
        imageName = "iceCrystal",
        width = 60, 
        height = 60,
        place = "sky"
        }, {
        name = "dead tree",
        color = {r=195, g=60, b=60},
        imageName = "deadTree",
        width = 20, 
        height = 100,
        place = "floor"
    },
}


function PlayStateConstructor.new(o)
    o = o or {}
    setmetatable(o, mt)    
    return o
end

function PlayState:init(assetManager)
    self.smallFont = assetManager:getFont("aesymatt11")
    self.font = assetManager:getFont("aesymatt14")
    self.bigFont = assetManager:getFont("aesymatt28")
    self.dieSound = assetManager:getSound("die")
    self.hitSound = assetManager:getSound("hit")
    self.startSound = assetManager:getSound("gameStart")
    
    self.backdropTreeImage = assetManager:getImage("backdropTree")
    --self.backdropCloudImage = assetManager:getImage("backdropCloud")
    for _,v in ipairs(hazardTypes) do
        v.image = assetManager:getImage(v.imageName)
    end   
    
    self.music = assetManager:getMusic("gameMusic")
    playMusic(self.music)   
    
    self.assets = assetManager
    
    self:reset()
end

function PlayState:reset()
    local screenWidth = love.graphics.getWidth() / GLOBAL_SCALE
    local screenHeight = love.graphics.getHeight() / GLOBAL_SCALE

    self.stateTimer = 0
    self.gameOver = false
    self.floorY = screenHeight/2 + 60
    self.player = Player.new({}, 100, self.floorY, self.assets)

    self.score = 0
    self.scoreSinceLastSettlement = 0
    self.distance = 0
    self.warmth = 100
    self.allowHazardSpawn = true
    self.hazardSpawnDistance = 1
    self.settlementsReached = 0
    self.lastSettlementAt = SETTLEMENT_SPACING_BASE
    self.nextSettlementAt = SETTLEMENT_SPACING_BASE
    self.lastHitBy = "nothing"

    local blockWidth = 40
    local blockHeight = 100
    local blockNum = 10
    local spacing = (screenWidth - (blockWidth*(blockNum-2))) / blockNum

    self.backdropTreePool = {}
    for i=0,blockNum-1 do 
        self.backdropTreePool[#self.backdropTreePool + 1] = {
            x=i * (blockWidth + spacing), 
            y=self.floorY - blockHeight,
            width=blockWidth,
            height=blockHeight            
        }
    end

    self.hazards = {}

    self.backgroundSnow = SnowParticles.new({scrollFactor=0})
    self.foregroundSnow = SnowParticles.new({scrollFactor=1})
    
    self.playerFire = FireParticles.new({x = self.player.x, y = self.player.y})
    
    playSound(self.startSound, 0.1, {x = self.player.x, y = self.player.y})
    
    self.warmthBar = StatusBar.new({x = screenWidth/2 - 100, 
            y = screenHeight - 50, 
            color = {255, 128, 0}, 
            name = "Warmth",
            status = self.warmth,
            font = self.font,
            width = 200,
            height = 20,
        })
    
    self.speedBar = StatusBar.new({x = screenWidth/2 - 100, 
            y = screenHeight - 20, 
            color = {0, 128, 255}, 
            name = "Speed",
            status = MAX_SPEED - MIN_SPEED,
            font = self.font,
            width = 200,
            height = 10,
            })
    self.speedBar.status = MIN_SPEED
    
    --love.audio.setPosition(self.player.x/10, self.player.y/10, 0)
end

function PlayState:update(dt)

    local screenWidth = love.graphics.getWidth() / GLOBAL_SCALE
    local screenHeight = love.graphics.getHeight() / GLOBAL_SCALE
    
    self.warmthBar:update(dt, self.warmth)
    self.speedBar:update(dt, self.player.groundSpeed - MIN_SPEED)

    local movement = self.player.totalGroundSpeed * dt
    if self.gameOver then
        self.slowdownFactor = self.slowdownFactor * 0.9
        movement = movement * self.slowdownFactor
    end
    
    self.player:update(dt, self.floorY, self.stateTimer*2)
    self.playerFire.x = self.player.x - 10
    self.playerFire.y = self.player.y - 5

    self.backgroundSnow:update(dt, movement)
    self.foregroundSnow:update(dt, movement)
    self.playerFire:update(dt, movement)

    if self.settlement and self.settlement.alive then
        self.settlement:update(dt, movement, self.player.x)
        local torchScoreBonus = self.settlement:checkTorches(self.player.x)
        if torchScoreBonus > 0 then
            self.warmth = 100
        end            
        self.score = self.score + torchScoreBonus
    end

    if not self.gameOver then
        self.stateTimer = self.stateTimer + dt  
        
        -- update hazards, and check for collision against the player
        for i,v in ipairs(self.hazards) do
            if v.alive then
                v.x = v.x - self.player.totalGroundSpeed * dt
                -- did it go offscreen?
                if v.x < -v.width then
                    -- kill the hazard    
                    v.alive = false
                end

                if not v.hit then
                    if checkEntityCollision(self.player, v) then
                        v.hit = true
                        self.warmth = self.warmth - 10
                        playSound(self.hitSound, 0.3)
                        self.lastHitBy = v.name
                    end
                end
            end
        end

        -- update art elements
        for _,v in ipairs(self.backdropTreePool) do
            v.x = v.x - self.player.totalGroundSpeed * dt
            if(v.x < -v.width) then
                v.x = screenWidth + v.width
            end
        end

        -- add distance reached to score
        local scoreIncrease = (movement / 10)
        self.score = self.score + scoreIncrease
        self.scoreSinceLastSettlement = self.scoreSinceLastSettlement + scoreIncrease
        self.distance = self.distance + (movement / 10)
        local difficulty = (1 - self.scoreSinceLastSettlement/10000) * 30 + 30


        -- settlements
        -- allow for a gap either side of a settlement spawn to remove spawns
        if self.distance > self.nextSettlementAt - SETTLEMENT_SAFE_GAP then
            self.allowHazardSpawn = false
        end
        if ((self.settlement and not self.settlement.alive) or not self.settlement) and self.distance > self.nextSettlementAt then
            -- spawn settlement
            self.settlement = Settlement.new({x = screenWidth + SETTLEMENT_WIDTH/2, y = self.floorY}, self.assets) 
            self.lastSettlementAt = self.nextSettlementAt
            self.nextSettlementAt = self.lastSettlementAt + SETTLEMENT_SPACING_BASE + self.settlementsReached * SETTLEMENT_SPACING_INCREASE
            self.scoreSinceLastSettlement = 0
            self.settlementsReached = self.settlementsReached + 1
        end
        -- turn hazard spawning back on after clearing the settlement
        if not self.allowHazardSpawn and self.distance > self.lastSettlementAt + SETTLEMENT_SAFE_GAP*2 then
            self.allowHazardSpawn = true
        end






        -- spawn hazards
        if self.distance > self.hazardSpawnDistance and self.allowHazardSpawn then
            local hazardType = hazardTypes[love.math.random(1, #hazardTypes)]
            --print("spawned hazard of type " .. hazardType.name)
            local hazardY = 0
            if hazardType.place == "floor" then
                hazardY = self.floorY
            elseif hazardType.place == "sky" then
                hazardY = (screenHeight - self.floorY) / 2 + math.random(-20, 20)
            end

            local hazard = nil
            -- look for an existing dead instance to revive
            for _,v in ipairs(self.hazards) do
                if v.name == hazardType.name and not v.alive and hazard == nil then
                    v.x = screenWidth + hazardType.width
                    v.y = hazardY + (hazardType.height/2 * (hazardType.place == "floor" and -1 or 1))
                    v.alive = true
                    v.hit = false
                    v.image = hazardType.image
                    hazard = v
                end            
            end

            if not hazard then
                -- we need to make a new hazard table and put it into the relevant tables
                hazard = {
                    name = hazardType.name,
                    color = hazardType.color,
                    width = hazardType.width,
                    height = hazardType.height,
                    display = hazardType.display,
                    x = screenWidth + hazardType.width,
                    y = hazardY + (hazardType.height/2 * (hazardType.place == "floor" and -1 or 1)),
                    alive = true,
                    hit = false,
                    image = hazardType.image
                }
                self.hazards[#self.hazards+1] = hazard
            end

            self.hazardSpawnDistance = self.distance + love.math.random(difficulty/2, difficulty)
        end

        self.warmth = self.warmth - dt
        if self.warmth <= 0 then
            self.warmth = 0      
            self.playerFire.alive = false
            self.gameOver = true
            self.slowdownFactor = 1
            self.player.vel.x = self.player.totalGroundSpeed
            self.player.alive = false
            playSound(self.dieSound, 0.1)
        end

    end


    -- shader animation timer
--    self.shader:send("time", self.shaderTimer)

end

function PlayState:keypressed(key, scancode, isrepeat)
    if not self.gameOver then
        if key == "space" or key == "x" or key == "up" or key == "w" then
            self.player:jump()
        elseif key == "c" or key == "down" or key == "s" then
            self.player:dive()
        end

        if key == "left" or key == "a" then
            self.player:beginAccel(-1)
        elseif key == "right" or key == "d" then
            self.player:beginAccel(1)
        end
    else
        if key == "r" then
            self:reset()
        elseif key == "s" then
            changeState("MenuState")
        end
    end
    if key == "m" then
        toggleSound()
    elseif key == "0" or key == "o" then
        toggleMusic()                
    end
end

function PlayState:keyreleased(key)
    if not self.gameOver then
        if key == "space" or key == "x" then
            self.player:endJump()
        end

        if key == "left" or key == "right" or key == "a" or key == "d" then
            self.player:endAccel()
        end
    end
end

function PlayState:draw()

    love.graphics.scale(GLOBAL_SCALE, GLOBAL_SCALE)
    local screenWidth = love.graphics.getWidth() / GLOBAL_SCALE
    local screenHeight = love.graphics.getHeight() / GLOBAL_SCALE

    -- draw the sky
    love.graphics.setColor(54, 52, 72)
    love.graphics.rectangle("fill", 0, 0, screenWidth, screenHeight)

    -- draw background snow
    self.backgroundSnow:draw()

    -- draw the ground
    love.graphics.setColor(90, 100, 110)
    love.graphics.rectangle("fill", 0, self.floorY, screenWidth, screenHeight - self.floorY)

    -- draw background
    love.graphics.setColor(255, 255, 255)
    for _,v in ipairs(self.backdropTreePool) do
        love.graphics.draw(self.backdropTreeImage, v.x, v.y)
    end

    -- draw settlements
    if self.settlement and self.settlement.alive then
        self.settlement:draw()
    end


    -- draw hazards
    for _,v in ipairs(self.hazards) do
        if v.alive then
            --love.graphics.setColor(v.color.r, v.color.g, v.color.b)
            --love.graphics.rectangle("fill", v.x - v.width/2, v.y - v.height/2, v.width, v.height)
            if v.hit then
                love.graphics.setColor(128, 128, 128)
            else
                love.graphics.setColor(255, 255, 255)
            end            
            love.graphics.draw(v.image, v.x - v.width/2, v.y - v.height/2)
        end
    end    

    -- draw the player
    self.player:draw()


    -- draw fire
    self.playerFire:draw()

    -- draw foreground snow
    self.foregroundSnow:draw()

    -- draw ui
    --love.graphics.scale(1, 1)
    love.graphics.setColor(255, 255, 255)
    
    self.warmthBar:draw()
    self.speedBar:draw()
    
    love.graphics.setFont(self.bigFont)
    love.graphics.setColor(128, 128, 32)
    love.graphics.printf(math.floor(self.score), 1, 11, screenWidth - 1, "center")
    love.graphics.setColor(255, 255, 64)
    love.graphics.printf(math.floor(self.score), 0, 10, screenWidth, "center")

    if self.gameOver then
        love.graphics.setFont(self.font)
        local plural = self.settlementsReached == 1 and "" or "s"
        local gameOverMessage = {{255, 255, 255}, "- GAME OVER -\n\n\nYou were frozen by a ", 
            {0, 128, 255}, self.lastHitBy, 
            {255, 255, 255}, ".\nYou brought fire to ",
            {255, 128, 0}, self.settlementsReached,
            {255, 255, 255}, " settlement" .. plural .. ".\nYou ran ", 
            {128, 128, 128}, math.floor(self.distance)/1000,
            {255, 255, 255}, "m.\nYour final score is ",
            {255, 255, 0}, math.floor(self.score),
            {255, 255, 255}, ".\n\n(R) to retry.\n(S) for start menu."}
        love.graphics.setColor(0, 0, 0)
        love.graphics.printf(gameOverMessage, 
            0, screenHeight/2 - 69, screenWidth - 1, "center")
        love.graphics.setColor(255, 255, 255)
        love.graphics.printf(gameOverMessage, 
            0, screenHeight/2 - 70, screenWidth, "center")
    end

    love.graphics.setColor(255, 255, 255)
    love.graphics.setFont(self.smallFont)
    if not ENABLE_SOUND then
        love.graphics.print("Sound muted. (M)", 6, screenHeight - 30)
    end
    if not PLAY_MUSIC then
        love.graphics.print("Music muted. (0)", 6, screenHeight - 15)
    end    

    -- background effects
--    love.graphics.setShader(self.shader)
--    love.graphics.rectangle("fill", 0, 0, screenWidth, screenHeight)
--    love.graphics.setShader()

end

function PlayState:resize(w, h)
    --self.shader:send("resolution", { w, h })
end

return PlayStateConstructor